import 'package:utilary/utilary.dart';
import 'chessman.dart';
import 'chess_board.dart';
import 'move.dart';

///A notation that allows one to represent the state of
///a chess board using alphanumeric characters.
///Pieces are represented by their letter symbols (E.G Q=Queen, N=Knight).
///Capital letters are for white chessman, lowercase letters for black ones.
///Empy squares are represented by numbers (E.G. 3=three empty space).
///The position is recorded rank by rank starting with the eight rank (a8)
///the ranks are separeted by a '/'
class ForsythNotation {
  static const String startingBoard =
      "rnbqkbnr/pppppppp/8/8/8/8/pppppppp/RNBQKBNR";

  static String toForsyth(ChessBoard board) {
    String result = "";

    for (int rank = 7; rank >= 0; rank--) {
      Chessman piece;
      int emptyTiles = 0;
      for (int file = 0; file < 8; file++) {
        piece = board.get(file, rank);

        if (piece == null) {
          emptyTiles++;
        } else {
          if (emptyTiles > 0) {
            result += emptyTiles.toString();
            emptyTiles = 0;
          }
          result += piece.forsythChar;
        }
      }

      if (emptyTiles > 0) {
        result += emptyTiles.toString();
      }

      //end of rand
      if (rank > 0) {
        result += '/';
      }
    }

    return result;
  }

  static ChessBoard fromForsyth(String state) {
    ChessBoard board = ChessBoard.emptyBoard();

    int rank = 7;
    int file = 0;
    String char;
    Chessman piece;
    for (int i = 0; i < state.length; i++) {
      char = state[i];

      //char is number
      if ("123456789".contains(char)) {
        file += int.parse(char) - 1; //skip char file
      } else if (char == '/') {
        rank--; //start of new rank
        file = 0;
        continue;
      } else if ("kK".contains(char)) {
        piece = King.fromForsythChar(board, char, Vector(file, rank));
      } else if ("qQ".contains(char)) {
        piece = Queen.fromForsythChar(board, char, Vector(file, rank));
      } else if ("bB".contains(char)) {
        piece = Bishop.fromForsythChar(board, char, Vector(file, rank));
      } else if ("nN".contains(char)) {
        piece = Knight.fromForsythChar(board, char, Vector(file, rank));
      } else if ("rR".contains(char)) {
        piece = Rook.fromForsythChar(board, char, Vector(file, rank));
      } else if ("pP".contains(char)) {
        piece = Pawn.fromForsythChar(board, char, Vector(file, rank));
      }

      if (piece != null) board.add(piece);
      file++;
    }

    return board;
  }
}

///Used to represent moves
class ArithmeticNotation {
  static int fileToIndex(String char) => char.codeUnitAt(0) - 'a'.codeUnitAt(0);
  static String indexToFile(int index) =>
      String.fromCharCode('a'.codeUnitAt(0) + index);

  static int rankToIndex(int rank) => rank - 1;
  static int indexToRank(int index) => index + 1;

  static Vector algebraicPosToVector(String pos) {
    int x = fileToIndex(pos[0]);
    int y = rankToIndex(int.parse(pos[1]));

    return Vector(x, y);
  }

  static String VectorToAlgebraicPos(Vector pos) {
    String file = indexToFile(pos.x);
    String rank = indexToRank(pos.y).toString();

    return file + rank;
  }

  final ChessBoard board;
  final Chessman actor;
  final Vector to;

  final CastleType castling;

  factory ArithmeticNotation(ChessBoard board, String text) {
    //Check for castling
    if (text == "O-O-O") {
      return ArithmeticNotation._(
          board, board.king, null, CastleType.queenSide);
    }
    if (text == "O-O") {
      return ArithmeticNotation._(board, board.king, null, CastleType.kingSide);
    }

    //Remove unwanted characters like capture 'x', check '+', check-mate '#'
    //to obtain a move string that consists of only source, actor-chessman, destination
    text.replaceFirst(r"[x+#]", '');

    //Identify destination (last 2 characters)
    Vector destination = algebraicPosToVector(text.substring(text.length - 2));
    //Remove destination from text
    text = text.substring(0, text.length - 2);

    //Everything remaining now is optional:
    //Identify source position
    int rank, file;
    for (int i = 0; i < 2; i++) {
      String char = text[text.length - 1 - i];

      if ("abcdefgh".contains(char)) {
        file = fileToIndex(char);
      } else if ("12345678".contains(char)) {
        rank = rankToIndex(int.parse(char));
      } else {
        break;
      }
    }
    //remove source from text
    int charCount = 0;
    if (rank != null && file != null)
      charCount = 2;
    else if (rank != null || file != null) charCount = 1;

    text = text.substring(0, text.length - charCount);

    //Identify actor
    String actorChar;
    List<Chessman> possibleActors;

    if (text.length > 0) actorChar = text[0];

    actorChar ??= 'P';

    possibleActors = board.playerPieces
        .where((Chessman piece) => piece.char == actorChar)
        .toList();

    if (file != null) {
      possibleActors = possibleActors
          .where((Chessman piece) => piece.pos.x == file)
          .toList();
    }

    if (rank != null) {
      possibleActors = possibleActors
          .where((Chessman piece) => piece.pos.y == rank)
          .toList();
    }

    Chessman actor = possibleActors.first;

    return ArithmeticNotation._(board, actor, destination);
  }

  ArithmeticNotation._(this.board, this.actor, this.to, [this.castling]);

  Move toMove() =>
      castling == null ? Move(board, actor, to) : Castling(board, castling);
}
