import 'package:utilary/utilary.dart' show Vector, Direction;
import 'chess_board.dart';
import 'move.dart';

enum ChessColor { white, black }

///A single chess piece
abstract class Chessman {
  ChessBoard board;

  String _char;
  String get char => _char;
  set char(String char) => _char = char;

  String get forsythChar => isBlack ? _char.toLowerCase() : _char;

  final ChessColor color;
  bool get isWhite => color == ChessColor.white;
  bool get isBlack => color == ChessColor.black;

  bool isAlly(Chessman piece) => piece.color == color;
  bool isEnemy(Chessman piece) => !isAlly(piece);

  ChessColor get opponentsColor =>
      isWhite ? ChessColor.black : ChessColor.white;

  Vector _pos;
  Vector get pos => _pos;

  Chessman(this.board, this._pos, this.color);

  void move(Vector destination) => _pos = destination;

  bool get isInDanger => board.canBeAttacked(pos, opponentsColor);

  bool canMoveTo(Vector destination);

  bool operator ==(Object o) =>
      o is Chessman && char == o.char && pos == o.pos && color == o.color;

  int get hashCode => char.hashCode ^ _pos.hashCode;
}

class King extends Chessman {
  bool neverMoved = true;

  King(ChessBoard board, Vector pos, ChessColor color)
      : super(board, pos, color) {
    char = 'K';
  }

  King.fromForsythChar(ChessBoard board, String char, [Vector pos])
      : super(board, pos, char == 'K' ? ChessColor.white : ChessColor.black) {
    super.char = char.toUpperCase();
  }

  ///Returns `true` when these conditions are verified simultaniously:
  ///+ there isn't any piece amoung the enemy's that can attack [destination];
  ///+ [destination] is one step away from current [pos];
  ///+ no ally chessman at [destination];
  bool canMoveTo(Vector destination) =>
      !board
          .getPieces(opponentsColor)
          .any((Chessman piece) => piece.canMoveTo(destination)) &&
      (destination - pos).kingLength == 1 &&
      !board[destination].isAlly(this);
}

class Queen extends Chessman {
  Queen(ChessBoard board, Vector pos, ChessColor color)
      : super(board, pos, color) {
    char = 'Q';
  }

  Queen.fromForsythChar(ChessBoard board, String char, [Vector pos])
      : super(board, pos, char == 'Q' ? ChessColor.white : ChessColor.black) {
    super.char = char.toUpperCase();
  }

  ///Returns `true` when these condition are verified simultaniously:
  ///+ [destination] is on any of the main directions from [pos]
  ///+ no other chessman is blocking the path to [destination]
  bool canMoveTo(Vector destination) {
    Vector offset = (destination - pos);

    //Check if destination isn't on a main direction from pos
    if (offset.kingLength != offset.rookLength &&
        offset.x.abs() != offset.y.abs()) return false;

    //Get direction to destination
    Direction dir = offset.nearestDirection;
    //Check if any chessman is blocking the way to [destination]
    for (Vector tempPos = pos; tempPos != destination; tempPos += dir) {
      if (board[tempPos] != null) return false;
    }

    //No chessman is blocking the way to [destination]:
    //this piece can move there.
    return true;
  }
}

class Bishop extends Chessman {
  Bishop(ChessBoard board, Vector pos, ChessColor color)
      : super(board, pos, color) {
    char = 'B';
  }

  Bishop.fromForsythChar(ChessBoard board, String char, [Vector pos])
      : super(board, pos, char == 'B' ? ChessColor.white : ChessColor.black) {
    super.char = char.toUpperCase();
  }

  ///Returns `true` when these conditions are verified simultaniously:
  ///+ [destination] is on any of the intercardinal directions from [pos];
  ///+ no other chessman is blocking the way to [destination];
  bool canMoveTo(Vector destination) {
    Vector offset = (destination - pos);

    //Check if destination isn't on a main direction from pos
    if (offset.x.abs() != offset.y.abs()) return false;

    //Get direction to destination
    Direction dir = offset.nearestDirection;
    //Check if any chessman is blocking the way to [destination]
    for (Vector tempPos = pos; tempPos != destination; tempPos += dir) {
      if (board[tempPos] != null) return false;
    }

    //No chessman is blocking the way to [destination]:
    //this piece can move there.
    return true;
  }
}

class Knight extends Chessman {
  static const List<Vector> _movementPattern = [
    Vector(1, 2),
    Vector(1, -2),
    Vector(-1, 2),
    Vector(-1, -2),
    Vector(2, 1),
    Vector(-2, 1),
    Vector(2, -1),
    Vector(-2, -1)
  ];

  Knight(ChessBoard board, Vector pos, ChessColor color)
      : super(board, pos, color) {
    char = 'N';
  }

  Knight.fromForsythChar(ChessBoard board, String char, [Vector pos])
      : super(board, pos, char == 'N' ? ChessColor.white : ChessColor.black) {
    super.char = char.toUpperCase();
  }

  ///Returns `true` when these conditions are verified simultaniously:
  ///+ [destination] isn't occupied by an ally;
  ///+ The relative distance to [destination] from [pos] match [_movementPattern];
  bool canMoveTo(Vector destination) {
    Chessman piece = board[destination];

    bool validDestination = _movementPattern.contains(destination - pos);

    return validDestination && (piece != null ? isAlly(piece) : true);
  }
}

class Rook extends Chessman {
  bool neverMoved = true;

  Rook(ChessBoard board, Vector pos, ChessColor color)
      : super(board, pos, color) {
    char = 'R';
  }

  Rook.fromForsythChar(ChessBoard board, String char, [Vector pos])
      : super(board, pos, char == 'R' ? ChessColor.white : ChessColor.black) {
    super.char = char.toUpperCase();
  }

  void move(Vector destination) {
    super.move(destination);

    neverMoved = false;
  }

  ///Returns `true` when these condition are verified simultaniously:
  ///+ [destination] is on any of the main directions from [pos]
  ///+ no other chessman is blocking the path to [destination]
  bool canMoveTo(Vector destination) {
    Vector offset = (destination - pos);

    //Check if destination isn't on a main direction from pos
    if (offset.kingLength != offset.rookLength) return false;

    //Get direction to destination
    Direction dir = offset.nearestDirection;
    //Check if any chessman is blocking the way to [destination]
    for (Vector tempPos = pos; tempPos != destination; tempPos += dir) {
      if (board[tempPos] != null) return false;
    }

    //No chessman is blocking the way to [destination]:
    //this piece can move there.
    return true;
  }
}

class Pawn extends Chessman {
  bool neverMoved = true;
  bool justLongMoved = false;

  Pawn(ChessBoard board, Vector pos, ChessColor color)
      : super(board, pos, color) {
    char = 'P';
  }

  Pawn.fromForsythChar(ChessBoard board, String char, [Vector pos])
      : super(board, pos, char == 'P' ? ChessColor.white : ChessColor.black) {
    super.char = char.toUpperCase();
  }

  void move(Vector destination) {
    final Vector tempPos = pos;
    super.move(destination);

    neverMoved = false;
    justLongMoved = (destination - tempPos).rookLength == 2;
  }

  ///Returns `true` when these conditions are verified simultaniously:
  ///+ if [destination] isn't occupied:
  ///  * [destination] is 1 tile ahead of [pos] OR at most 2 tile if [nevermoved] is true
  ///    OR 1 tile behind a 2 tile moved enemy Pawn (en passant);
  ///+ if [destination] is occupied:
  ///  * Chessman at [destination] isn't an ally;
  ///  * [destination] is located at north-east or north-west of [pos];
  bool canMoveTo(Vector destination) {
    //relative north
    final Direction dir = isWhite ? Direction.north : Direction.south;
    //chessman at [destination]
    final Chessman piece = board[destination];

    if (piece == null) {
      final int maxDistance = neverMoved ? 2 : 1;
      final Vector offset = destination - pos;

      if (offset.kingLength > maxDistance) return false;

      //Check if 2 tile movement is permitted
      if (offset.kingLength == 2 && board[pos + dir] != null) return false;

      //Check en passant capture:
      //if [destination] is at relative north-east or north-west
      if ([dir.offsetX(1), dir.offsetX(-1)].contains(offset)) {
        Chessman piece = board[offset - dir];
        //if the piece behind destination is an enemy Pawn
        if (isEnemy(piece) && piece is Pawn) {
          //Return `true` if piece just moved 2 tile ahead
          return piece.justLongMoved;
        }

        return false;
      }

      return true;
    }

    //Is a capture move
    if ([pos + dir.offsetX(1), pos + dir.offsetX(-1)].contains(destination)) {
      return isEnemy(piece);
    }

    return false;
  }
}
