import 'package:utilary/utilary.dart';
import 'chessman.dart';
import 'chess_board.dart';
import 'chess_notation.dart';

enum CastleType { kingSide, queenSide }

class Move {
  final ChessBoard board;
  final Chessman actor;
  final Vector source;
  final Vector destination;

  Move(this.board, this.actor, this.destination) : source = actor.pos;

  MoveResult perform() {
    bool success = false;
    bool capture = false;

    if (actor.canMoveTo(destination)) {
      success = true;

      capture = board[destination] != null;
      board.move(actor, destination);
    }

    bool check = board.enemyKing.isInDanger;
    //identify check-mate
    bool checkMate = board.isCheckMate(board.enemy);

    return MoveResult(success,
        capture: capture, check: check, checkMate: checkMate);
  }

  String toString() =>
      "${actor.char + ArithmeticNotation.VectorToAlgebraicPos(destination)}";
}

class Castling extends Move {
  final CastleType side;

  Castling(ChessBoard board, this.side) : super(board, board.king, null);

  ///One cannot castle from, through or into check!
  MoveResult perform() {
    final King king = actor;
    if (king.isInDanger || !king.neverMoved) return MoveResult.failure;

    Direction dir = Direction.east;
    if (side == CastleType.queenSide) dir.rotate180Left();

    int rookY = actor.isWhite ? 0 : 7;
    int rookX = side == CastleType.kingSide ? 7 : 0;

    final Rook rook = board.get(rookX, rookY);
    if (rook == null || !rook.neverMoved) return MoveResult.failure;

    for (Vector pos in Line(king.pos + dir, rook.pos - dir)) {
      if (board.canBeAttacked(pos, board.enemy) || board[pos] != null) return MoveResult.failure;
    }

    board.move(king, king.pos + dir * 2);
    board.move(rook, actor.pos - dir);

    return MoveResult.success;
  }

  String toString() => side == CastleType.kingSide ? "O-O" : "O-O-O";
}

class MoveResult {
  static final MoveResult success = MoveResult(true);
  static final MoveResult failure = MoveResult(false);

  final bool succeed;
  final bool capture;
  final bool check;
  final bool checkMate;

  const MoveResult(this.succeed,
      {bool capture = false, bool check = false, bool checkMate = false})
      : this.capture = capture,
        this.check = check,
        this.checkMate = checkMate;

  bool get failed => !succeed;
}
